@@ -99,25 +99,29 @@ fn sort_by<'a>(xs: &mut [Val], f: impl Fn(Val) -> ValRs<'a>) -> Result<(), Error
 
 /// Group an array by the given function.
 fn group_by<'a>(xs: Vec<Val>, f: impl Fn(Val) -> ValRs<'a>) -> ValR {
-    let mut err = None;
-    let mut yx = xs
+    let mut yx: Vec<(Vec<Val>, Val)> = xs
         .into_iter()
-        .map(|x| (run_if_ok(x.clone(), &mut err, &f), x))
-        .collect::<Vec<(Vec<Val>, Val)>>();
-    if let Some(err) = err {
-        return Err(err);
-    }
+        .map(|x| Ok((f(x.clone()).collect::<Result<_, _>>()?, x)))
+        .collect::<Result<_, _>>()?;
 
     yx.sort_by(|(y1, _), (y2, _)| y1.cmp(y2));
 
-    // TODO: do not use itertools here (to remove dependency)
-    use itertools::Itertools;
-    let grouped = yx
-        .into_iter()
-        .group_by(|(y, _)| y.clone())
-        .into_iter()
-        .map(|(_y, yxs)| Val::arr(yxs.map(|(_y, x)| x).collect()))
-        .collect();
+    let mut grouped = Vec::new();
+    let mut yx = yx.into_iter();
+    if let Some((mut group_y, first_x)) = yx.next() {
+        let mut group = Vec::from([first_x]);
+        for (y, x) in yx {
+            if group_y != y {
+                grouped.push(Val::arr(core::mem::take(&mut group)));
+                group_y = y;
+            }
+            group.push(x);
+        }
+        if !group.is_empty() {
+            grouped.push(Val::arr(group))
+        }
+    }
+
     Ok(Val::arr(grouped))
 }
 
@@ -175,6 +179,28 @@ fn split(s: &str, sep: &str) -> Vec<Val> {
     }
 }
 
+/// This implements a ~10x faster version of:
+/// ~~~ text
+/// def range($from; $to; $by): $from |
+///    if $by > 0 then while(. < $to; . + $by)
+///    else            while(. > $to; . + $by)
+///    end;
+/// ~~~
+fn range(mut from: ValR, to: Val, by: Val) -> impl Iterator<Item = ValR> {
+    let positive = by > Val::Int(0);
+    core::iter::from_fn(move || match from.clone() {
+        Ok(x) if (positive && x < to) || (!positive && x > to) => {
+            Some(core::mem::replace(&mut from, x + by.clone()))
+        }
+        Ok(_) => None,
+        e @ Err(_) => {
+            // return None as following value
+            from = Ok(to.clone());
+            Some(e)
+        }
+    })
+}
+
 fn strip<F>(s: &Rc<String>, other: &str, f: F) -> Rc<String>
 where
     F: for<'a> Fn(&'a str, &str) -> Option<&'a str>,
@@ -206,58 +232,69 @@ fn to_csv(vs: &[Val]) -> Result<String, Error> {
     Ok(vs.iter().map(fr).collect::<Result<Vec<_>, _>>()?.join(","))
 }
 
+fn once_with<'a, T>(f: impl FnOnce() -> T + 'a) -> Box<dyn Iterator<Item = T> + 'a> {
+    Box::new(core::iter::once_with(f))
+}
+
 const CORE_RUN: &[(&str, usize, RunPtr)] = &[
     ("inputs", 0, |_, cv| {
         Box::new(cv.0.inputs().map(|r| r.map_err(Error::str)))
     }),
-    ("length", 0, |_, cv| box_once(length(&cv.1))),
+    ("length", 0, |_, cv| once_with(move || length(&cv.1))),
     ("keys_unsorted", 0, |_, cv| {
-        box_once(cv.1.keys_unsorted().map(Val::arr))
+        once_with(move || cv.1.keys_unsorted().map(Val::arr))
+    }),
+    ("floor", 0, |_, cv| {
+        once_with(move || cv.1.round(|f| f.floor()))
+    }),
+    ("round", 0, |_, cv| {
+        once_with(move || cv.1.round(|f| f.round()))
+    }),
+    ("ceil", 0, |_, cv| {
+        once_with(move || cv.1.round(|f| f.ceil()))
     }),
-    ("floor", 0, |_, cv| box_once(cv.1.round(|f| f.floor()))),
-    ("round", 0, |_, cv| box_once(cv.1.round(|f| f.round()))),
-    ("ceil", 0, |_, cv| box_once(cv.1.round(|f| f.ceil()))),
     ("tojson", 0, |_, cv| {
-        box_once(Ok(Val::str(cv.1.to_string())))
+        once_with(move || Ok(Val::str(cv.1.to_string())))
     }),
     ("utf8bytelength", 0, |_, cv| {
-        then(cv.1.as_str(), |s| box_once(Ok(Val::Int(s.len() as isize))))
+        once_with(move || cv.1.as_str().map(|s| Val::Int(s.len() as isize)))
     }),
     ("explode", 0, |_, cv| {
-        box_once(cv.1.as_str().and_then(|s| Ok(Val::arr(explode(s)?))))
+        once_with(move || cv.1.as_str().and_then(|s| Ok(Val::arr(explode(s)?))))
     }),
     ("implode", 0, |_, cv| {
-        box_once(cv.1.as_arr().and_then(|a| Ok(Val::str(implode(a)?))))
+        once_with(move || cv.1.as_arr().and_then(|a| Ok(Val::str(implode(a)?))))
     }),
     ("ascii_downcase", 0, |_, cv| {
-        box_once(cv.1.mutate_str(|s| s.make_ascii_lowercase()))
+        once_with(move || cv.1.mutate_str(|s| s.make_ascii_lowercase()))
     }),
     ("ascii_upcase", 0, |_, cv| {
-        box_once(cv.1.mutate_str(|s| s.make_ascii_uppercase()))
+        once_with(move || cv.1.mutate_str(|s| s.make_ascii_uppercase()))
     }),
     ("reverse", 0, |_, cv| {
-        box_once(cv.1.mutate_arr(|a| a.reverse()))
+        once_with(move || cv.1.mutate_arr(|a| a.reverse()))
+    }),
+    ("sort", 0, |_, cv| {
+        once_with(move || cv.1.mutate_arr(|a| a.sort()))
     }),
-    ("sort", 0, |_, cv| box_once(cv.1.mutate_arr(|a| a.sort()))),
     ("sort_by", 1, |args, cv| {
-        box_once(cv.1.try_mutate_arr(|arr| sort_by(arr, |v| args.get(0).run((cv.0.clone(), v)))))
+        once_with(move || {
+            cv.1.try_mutate_arr(|arr| sort_by(arr, |v| args.get(0).run((cv.0.clone(), v))))
+        })
     }),
     ("group_by", 1, |args, cv| {
-        then(cv.1.into_arr().map(rc_unwrap_or_clone), |arr| {
-            box_once(group_by(arr, |v| args.get(0).run((cv.0.clone(), v))))
-        })
+        let group = move |arr| group_by(arr, |v| args.get(0).run((cv.0.clone(), v)));
+        once_with(move || cv.1.into_arr().map(rc_unwrap_or_clone).and_then(group))
     }),
     ("min_by", 1, |args, cv| {
-        let f = |v| args.get(0).run((cv.0.clone(), v));
-        then(cv.1.into_arr().map(rc_unwrap_or_clone), |arr| {
-            box_once(cmp_by(arr, f, |my, y| y < my))
-        })
+        let f = move |v| args.get(0).run((cv.0.clone(), v));
+        let cmp = move |arr| cmp_by(arr, f, |my, y| y < my);
+        once_with(move || cv.1.into_arr().map(rc_unwrap_or_clone).and_then(cmp))
     }),
     ("max_by", 1, |args, cv| {
-        let f = |v| args.get(0).run((cv.0.clone(), v));
-        then(cv.1.into_arr().map(rc_unwrap_or_clone), |arr| {
-            box_once(cmp_by(arr, f, |my, y| y >= my))
-        })
+        let f = move |v| args.get(0).run((cv.0.clone(), v));
+        let cmp = move |arr| cmp_by(arr, f, |my, y| y >= my);
+        once_with(move || cv.1.into_arr().map(rc_unwrap_or_clone).and_then(cmp))
     }),
     ("has", 1, |args, cv| {
         let keys = args.get(0).run(cv.clone());
@@ -272,41 +309,17 @@ const CORE_RUN: &[(&str, usize, RunPtr)] = &[
         Box::new(seps.map(move |sep| Ok(Val::arr(split(cv.1.as_str()?, sep?.as_str()?)))))
     }),
     ("first", 1, |args, cv| Box::new(args.get(0).run(cv).take(1))),
-    ("last", 1, |args, cv| {
-        let last = args.get(0).run(cv).try_fold(None, |_, x| Ok(Some(x?)));
-        then(last, |y| Box::new(y.map(Ok).into_iter()))
-    }),
     ("limit", 2, |args, cv| {
         let n = args.get(0).run(cv.clone()).map(|n| n?.as_int());
         let f = move |n| args.get(1).run(cv.clone()).take(n);
         let pos = |n: isize| n.try_into().unwrap_or(0usize);
         Box::new(n.flat_map(move |n| then(n, |n| Box::new(f(pos(n))))))
     }),
-    // This implements a ~10x faster version of:
-    // ~~~ text
-    // def range($from; $to; $by): $from |
-    //    if $by > 0 then while(. < $to; . + $by)
-    //    else            while(. > $to; . + $by)
-    //    end;
-    // ~~~
     ("range", 3, |args, cv| {
         let (from, to, by) = (args.get(0), args.get(1), args.get(2));
-        Box::new(from.cartesian3(to, by, cv).flat_map(|(mut from, to, by)| {
+        Box::new(from.cartesian3(to, by, cv).flat_map(|(from, to, by)| {
             let to_by = to.and_then(|to| Ok((to, by?)));
-            then(to_by, |(to, by)| {
-                let positive = by > Val::Int(0);
-                Box::new(core::iter::from_fn(move || match from.clone() {
-                    Ok(x) if (positive && x < to) || (!positive && x > to) => {
-                        Some(core::mem::replace(&mut from, x + by.clone()))
-                    }
-                    Ok(_) => None,
-                    e @ Err(_) => {
-                        // return None as following value
-                        from = Ok(to.clone());
-                        Some(e)
-                    }
-                }))
-            })
+            then(to_by, |(to, by)| Box::new(range(from, to, by)))
         }))
     }),
     ("startswith", 1, |args, cv| {
@@ -332,19 +345,24 @@ const CORE_RUN: &[(&str, usize, RunPtr)] = &[
         }))
     }),
     ("@text", 0, |_, cv| {
-        box_once(Ok(Val::str(cv.1.to_string_or_clone())))
+        once_with(move || Ok(Val::str(cv.1.to_string_or_clone())))
+    }),
+    ("@json", 0, |_, cv| {
+        once_with(move || Ok(Val::str(cv.1.to_string())))
     }),
-    ("@json", 0, |_, cv| box_once(Ok(Val::str(cv.1.to_string())))),
     ("@sh", 0, |_, cv| {
-        let vs = match &cv.1 {
-            Val::Arr(a) => Box::new(a.iter()),
-            v => box_once(v),
+        let to_shs = |v: &Val| -> Result<Vec<String>, _> {
+            match v {
+                Val::Arr(a) => Box::new(a.iter()),
+                v => box_once(v),
+            }
+            .map(to_sh)
+            .collect()
         };
-        let ss = vs.map(to_sh).collect::<Result<Vec<_>, _>>();
-        box_once(ss.map(|ss| Val::str(ss.join(" "))))
+        once_with(move || to_shs(&cv.1).map(|ss| Val::str(ss.join(" "))))
     }),
     ("@csv", 0, |_, cv| {
-        box_once(cv.1.as_arr().and_then(|a| to_csv(a)).map(Val::str))
+        once_with(move || cv.1.as_arr().and_then(|a| to_csv(a)).map(Val::str))
     }),
 ];
 
@@ -358,7 +376,13 @@ fn now() -> Result<f64, Error> {
 }
 
 #[cfg(feature = "std")]
-const STD: &[(&str, usize, RunPtr)] = &[("now", 0, |_, _| box_once(now().map(Val::Float)))];
+const STD: &[(&str, usize, RunPtr)] = &[
+    ("env", 0, |_, _| {
+        let vars = std::env::vars().map(|(k, v)| (Rc::new(k), Val::str(v)));
+        once_with(|| Ok(Val::obj(vars.collect())))
+    }),
+    ("now", 0, |_, _| once_with(|| now().map(Val::Float))),
+];
 
 #[cfg(feature = "parse_json")]
 /// Convert string to a single JSON value.
@@ -372,7 +396,7 @@ fn from_json(s: &str) -> ValR {
 
 #[cfg(feature = "parse_json")]
 const PARSE_JSON: &[(&str, usize, RunPtr)] = &[("fromjson", 0, |_, cv| {
-    box_once(cv.1.as_str().and_then(|s| from_json(s)))
+    once_with(move || cv.1.as_str().and_then(|s| from_json(s)))
 })];
 
 #[cfg(feature = "format")]
@@ -391,37 +415,33 @@ fn to_tsv(vs: &[Val]) -> Result<String, Error> {
 #[cfg(feature = "format")]
 const FORMAT: &[(&str, usize, RunPtr)] = &[
     ("@tsv", 0, |_, cv| {
-        box_once(cv.1.as_arr().and_then(|a| to_tsv(a)).map(Val::str))
+        once_with(move || cv.1.as_arr().and_then(|a| to_tsv(a)).map(Val::str))
     }),
     ("@html", 0, |_, cv| {
-        let s = cv.1.to_string_or_clone();
-        let patterns = ["<", ">", "&", "\'", "\""];
-        let replacements = ["&lt;", "&gt;", "&amp;", "&apos;", "&quot;"];
-        box_once(Ok(Val::str(replace(&s, &patterns, &replacements))))
+        let pats = ["<", ">", "&", "\'", "\""];
+        let reps = ["&lt;", "&gt;", "&amp;", "&apos;", "&quot;"];
+        once_with(move || Ok(Val::str(replace(&cv.1.to_string_or_clone(), &pats, &reps))))
     }),
     ("@uri", 0, |_, cv| {
-        box_once(Ok(Val::str(
-            urlencoding::encode(&cv.1.to_string_or_clone()).into_owned(),
-        )))
+        use urlencoding::encode;
+        once_with(move || Ok(Val::str(encode(&cv.1.to_string_or_clone()).into_owned())))
     }),
     ("@base64", 0, |_, cv| {
-        use base64::{engine::general_purpose, Engine as _};
-        box_once(Ok(Val::str(
-            general_purpose::STANDARD.encode(cv.1.to_string_or_clone()),
-        )))
+        use base64::{engine::general_purpose::STANDARD, Engine};
+        once_with(move || Ok(Val::str(STANDARD.encode(cv.1.to_string_or_clone()))))
     }),
     ("@base64d", 0, |_, cv| {
-        use base64::{engine::general_purpose, Engine as _};
-        box_once(
-            general_purpose::STANDARD
+        use base64::{engine::general_purpose::STANDARD, Engine};
+        once_with(move || {
+            STANDARD
                 .decode(cv.1.to_string_or_clone())
                 .map_err(Error::str)
                 .and_then(|d| {
                     std::str::from_utf8(&d)
                         .map_err(Error::str)
                         .map(|s| Val::str(s.to_owned()))
-                }),
-        )
+                })
+        })
     }),
 ];
 
@@ -520,10 +540,10 @@ const REGEX: &[(&str, usize, RunPtr)] = &[
 #[cfg(feature = "time")]
 const TIME: &[(&str, usize, RunPtr)] = &[
     ("fromdateiso8601", 0, |_, cv| {
-        then(cv.1.as_str(), |s| box_once(time::from_iso8601(s)))
+        once_with(move || cv.1.as_str().and_then(|s| time::from_iso8601(s)))
     }),
     ("todateiso8601", 0, |_, cv| {
-        box_once(time::to_iso8601(&cv.1).map(Val::str))
+        once_with(move || time::to_iso8601(&cv.1).map(Val::str))
     }),
 ];
 
@@ -552,6 +572,6 @@ fn debug<T: core::fmt::Display>(x: T) -> T {
 const LOG: &[(&str, usize, RunPtr, UpdatePtr)] = &[(
     "debug",
     0,
-    |_, cv| box_once(Ok(debug(cv.1))),
+    |_, cv| once_with(move || Ok(debug(cv.1))),
     |_, cv, f| f(debug(cv.1)),
 )];
