@@ -8,8 +8,14 @@ use jaq_syn::filter::FoldType;
 use jaq_syn::{MathOp, OrdOp};
 
 /// Function from a value to a stream of value results.
-#[derive(Debug, Default, Clone)]
-pub struct Owned(Id, Vec<Ast>);
+#[derive(Debug, Clone)]
+pub struct Owned(Id, Box<[Ast]>);
+
+impl Default for Owned {
+    fn default() -> Self {
+        Self(Id(0), Box::new([Ast::Id]))
+    }
+}
 
 #[derive(Debug, Copy, Clone)]
 pub struct Ref<'a>(Id, &'a [Ast]);
@@ -35,12 +41,12 @@ pub(crate) struct Call {
     pub id: Id,
     pub typ: CallTyp,
     pub skip: usize,
-    pub args: Vec<Bind<Id, Id>>,
+    pub args: Box<[Bind<Id, Id>]>,
 }
 
 impl Owned {
     pub(crate) fn new(main: Id, recs: Vec<Ast>) -> Self {
-        Self(main, recs)
+        Self(main, recs.into())
     }
 }
 
@@ -55,7 +61,7 @@ pub(crate) enum Ast {
     Float(f64),
     Str(String),
     Array(Id),
-    Object(Vec<(Id, Id)>),
+    Object(Box<[(Id, Id)]>),
 
     Try(Id, Id),
     Neg(Id),
@@ -101,7 +107,7 @@ pub(crate) enum Ast {
     Var(usize),
     Call(Call),
 
-    Native(Native, Vec<Id>),
+    Native(Native, Box<[Id]>),
 }
 
 // we can unfortunately not make a `Box<dyn ... + Clone>`
@@ -271,12 +277,18 @@ impl<'a> FilterT<'a> for Ref<'a> {
             Ast::Ite(if_, then_, else_) => w(if_).pipe(cv, move |cv, v| {
                 w(if v.as_bool() { then_ } else { else_ }).run(cv)
             }),
-            Ast::Path(f, path) => then(path.eval(|p| w(p).run(cv.clone())), |path| {
-                let outs = w(f).run(cv).map(move |i| path.collect(i?));
-                Box::new(
-                    outs.flat_map(|vals| then(vals, |vals| Box::new(vals.into_iter().map(Ok)))),
-                )
-            }),
+            Ast::Path(f, path) => {
+                let path = path.map_ref(|i| {
+                    let cv = cv.clone();
+                    crate::into_iter::collect_if_once(move || w(i).run(cv))
+                });
+                flat_map_with(w(f).run(cv), path, |y, path| {
+                    then(y, |y| {
+                        flat_map_with(path.explode(), y, |path, y| then(path, |path| path.run(y)))
+                    })
+                })
+            }
+
             Ast::Update(path, f) => w(path).update(
                 (cv.0.clone(), cv.1),
                 Box::new(move |v| w(f).run((cv.0.clone(), v))),
@@ -309,9 +321,9 @@ impl<'a> FilterT<'a> for Ref<'a> {
                 match typ {
                     FoldType::Reduce => Box::new(fold(false, xs, Output(init), f)),
                     FoldType::For => Box::new(fold(true, xs, Output(init), f)),
-                    FoldType::Foreach => Box::new(init.flat_map(move |i| {
-                        then(i, |i| Box::new(fold(true, xs.clone(), Input(i), f.clone())))
-                    })),
+                    FoldType::Foreach => flat_map_with(init, xs, move |i, xs| {
+                        then(i, |i| Box::new(fold(true, xs, Input(i), f.clone())))
+                    }),
                 }
             }
 
@@ -360,14 +372,17 @@ impl<'a> FilterT<'a> for Ref<'a> {
             Ast::Fold(..) => todo!(),
 
             Ast::Id => f(cv.1),
-            Ast::Path(l, path) => w(l).update(
-                (cv.0.clone(), cv.1),
-                Box::new(move |v| {
-                    then(path.eval(|i| w(i).run((cv.0.clone(), v.clone()))), |path| {
-                        path.update(v, &f)
-                    })
-                }),
-            ),
+            Ast::Path(l, path) => {
+                let path = path.map_ref(|i| {
+                    let cv = cv.clone();
+                    crate::into_iter::collect_if_once(move || w(i).run(cv))
+                });
+                let f = move |v| {
+                    let mut paths = path.clone().explode();
+                    box_once(paths.try_fold(v, |acc, path| path?.update(acc, &f)))
+                };
+                w(l).update(cv, Box::new(f))
+            }
             Ast::Pipe(l, false, r) => w(l).update(
                 (cv.0.clone(), cv.1),
                 Box::new(move |v| w(r).update((cv.0.clone(), v), f.clone())),
